// LIBRERIA CON LE FUNZIONI PER COLORARE IL TESTO SOTTO LINUX
//
// (C) CTIO & Simone Gasparini some rights reserved


#include "tcolor.h"
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <sys/ioctl.h>


// ######################################################## INIZIALIZZAZIONI VARIABILI GLOBALI
char a_control_char[256];
clrchar replacement_char = { 'X', CLR( CWHITE, CRED ) };

int vt_col, vt_row, tc_col, tc_row;

ct_image screen_buf = {0,0,0,0,0,0,0,{0,0},0};
ct_image screen_mem = {0,0,0,0,0,0,0,{0,0},0};
clrchar  scr_blank = { ' ', CLR( CLIGHTGRAY, CBLACK ) };
clr_t output_clr = 0;
int box_char_switch = 0;
char ansi_clr[] = "\033[A;3F;4Bm";
const size_t ansi_clr_length = 10;

#define AT_POS 2
#define FG_POS 5
#define BG_POS 8

int cursor_visible = 0;
int off_screen = 0;


// ######################################################## USCITE IN CASO DI ERRORI E INTERRUZIONI
static void graceful_exit( int sig ) {

   exit( sig );
}

void trap_signals() {

/* Deliberate Interruption */
   signal( SIGHUP,  graceful_exit );   /* Terminal disconnected */
   signal( SIGINT,  graceful_exit );   /* Ctrl-C */
   signal( SIGQUIT, graceful_exit );   /* Similar to Ctrl-C, different key */
   signal( SIGTERM, graceful_exit );   /* Polite form of SIGINT */
   signal( SIGKILL, graceful_exit );   /* Unblockable kill signal */

/* Program Errors */
   signal( SIGFPE,  graceful_exit );   /* Floating Point exception */
   signal( SIGILL,  graceful_exit );   /* Illegal instruction */
   signal( SIGSEGV, graceful_exit );   /* Segment violation (legal memory) */
   signal( SIGABRT, graceful_exit );   /* abort() called */
   signal( SIGBUS,  graceful_exit );   /* Segment violation (illegal memory) */
   signal( SIGIOT,  graceful_exit );   /* Same as SIGABRT on most systems */
   signal( SIGSYS,  graceful_exit );   /* Bad system call */
   signal( SIGSTKFLT, graceful_exit ); /* Stack Fault */
}

void restore_signals() {

/* Deliberate Interruption */
   signal( SIGHUP,  SIG_DFL );  
   signal( SIGINT,  SIG_DFL );  
   signal( SIGQUIT, SIG_DFL );  
   signal( SIGTERM, SIG_DFL );  
   signal( SIGKILL, SIG_DFL );  

/* Program Errors */
   signal( SIGFPE,  SIG_DFL );  
   signal( SIGILL,  SIG_DFL );  
   signal( SIGSEGV, SIG_DFL );  
   signal( SIGABRT, SIG_DFL );  
   signal( SIGBUS,  SIG_DFL );  
   signal( SIGIOT,  SIG_DFL );  
   signal( SIGSYS,  SIG_DFL );  
   signal( SIGSTKFLT, SIG_DFL );
}


// ########################################################

// INIZIALIZZAZIONI

int init_display();
void restore_display();


int init_ctio() { 

   int status = 0;

   status |= init_display();

   trap_signals();

   return status;
}

void uninit_ctio() { 

   restore_display();

   restore_signals();
}


// ######################################################## FUNZIONI PER INIZIALIZZARE IL MONITOR

inline void enable_box_chars() {
   fwrite( "\033[11m", sizeof(char), 5, stdout );
   box_char_switch = BOX_BIT;
   output_clr |= BOX_BIT;
}


inline void disable_box_chars() {
   fwrite( "\033[10m", sizeof(char), 5, stdout );
   box_char_switch = 0;
   output_clr &= ~BOX_BIT;
}

inline void set_output_colour( clr_t new_clr ) {

   ansi_clr[AT_POS] = new_clr & CT_BRIGHT? '1' : '0';
   ansi_clr[FG_POS] = '0' + ( new_clr & 7 );
   ansi_clr[BG_POS] = '0' + ( ( new_clr >> 4 ) & 7 );

   fwrite( ansi_clr, sizeof(char), ansi_clr_length, stdout );

   output_clr = new_clr;

   if( ( new_clr & BOX_BIT ) != box_char_switch ) {

      if( box_char_switch )
         disable_box_chars();
      else
         enable_box_chars();
      }
}

const char *ctio_error_str( int error ) {

   static char strbuf[128];
   strbuf[0] = '\0';

   if( error ) {
      strcat( strbuf, "Cannot init CTIO because I " );
      if( error & CANNOT_INIT_KBD ) strcat( strbuf, "CANNOT_INIT_KBD " );
      if( error & CANNOT_INIT_MOUSE ) strcat( strbuf, "CANNOT_INIT_MOUSE" );
      if( error & CANNOT_INIT_DISPLAY ) strcat( strbuf, "CANNOT_INIT_DISPLAY " );
      strcat( strbuf, "\n" );
      }

   return strbuf;
}

inline void put_vt_cursor( int col, int row ) {

   printf( "\033[%d;%dH", row+1, col+1 );
   vt_col = col; vt_row = row; 
}

int ct_set_scroll_region( ct_image *image, int top, int bottom ) {

   if( !image || top < 0 || bottom < top || bottom >= image->h )
      return -1;

   image->scroll_top = top;
   image->scroll_bottom = bottom;
   return 0;
}

int vt_set_scroll_region( int top, int bottom ) {

   int error = ct_set_scroll_region( &screen_buf, top, bottom );
   error |= ct_set_scroll_region( &screen_mem, top, bottom );

   if( !error )
      printf( "\033[%d;%dr", screen_buf.scroll_top+1, 
                             screen_buf.scroll_bottom+1 );
   return error;
}

inline void show_vt_cursor() {

   fwrite( "\033[?25h\033[?0c", sizeof(char), 11, stdout );
}

void restore_display() {

   if( screen_buf.img ) { free( screen_buf.img ); screen_buf.img = 0; }
   if( screen_mem.img ) { free( screen_mem.img ); screen_mem.img = 0; }

   disable_box_chars();
   vt_set_scroll_region( 0, screen_buf.h - 1 );
   set_output_colour( CLR( CLIGHTGRAY, CBLACK ) );
   put_vt_cursor( 0, screen_buf.h-1 );
   show_vt_cursor();
}

int ct_fill( ct_image *image, clrchar blank ) {

   int cell;

   if( !image ) return -1;

   for( cell = 0; cell < image->size; cell++ )
      image->img[cell] = blank;

   return 0;
}

int ct_clear( ct_image *image ) {

   return ct_fill( image, image->blank );
}

ct_image ct_create( int x, int y, int w, int h, clrchar filler ) {

   ct_image image = {0,0,0,0,0,0,0,{0,0},0};

   if( w > 0 && h > 0 )
      image.img = (clrchar*)malloc( w * h * sizeof( clrchar ) );

   if( image.img ) {
      image.x = x;
      image.y = y;
      image.w = w;
      image.h = h;
      image.size = w * h;
      image.scroll_top = 0;
      image.scroll_bottom = image.h - 1;
      image.blank = filler;
      ct_clear( &image );
      }

   return image;
}

void prepare_control_char_lookup() {

   memset( a_control_char, 0, sizeof(a_control_char) );
   memset( a_control_char, 2, 32 );
   memset( a_control_char+127, 2, 33 );
   a_control_char[0] = 1;
   a_control_char[8] = 1;
   a_control_char[10] = 1;
   a_control_char[12] = 1;
   a_control_char[13] = 1;
   a_control_char[14] = 1;
   a_control_char[15] = 1;
   a_control_char[27] = 1;
   a_control_char[155] = 1;
}

int get_screen_size( int *cols, int *rows ) {

   struct winsize win;

   *cols = 80;
   *rows = 25;

   if( ioctl( fileno( stdin ), TIOCGWINSZ, &win ) == 0 ) {
      *cols = win.ws_col;
      *rows = win.ws_row;
      return 0;
      }

   return -1;
}

int set_screen_size( int cols, int rows ) {

   clrchar mem_blank = { ~0, ~0 };
   int size = rows * cols;

   if ( size < 0 ) return CANNOT_INIT_DISPLAY;

   if( screen_buf.img ) free( screen_buf.img );
   if( screen_mem.img ) free( screen_mem.img );

/* The buffers are made bigger than necessary, so that a stop
 * byte can be added at the end. This improves the loop logic
 * in vt_refresh_section(), so refresh is quicker.
 */
   screen_buf = ct_create( 0, 0, cols, rows + 1, scr_blank );
   screen_mem = ct_create( 0, 0, cols, rows + 1, mem_blank );

   if( !screen_buf.img || !screen_mem.img )
      return CANNOT_INIT_DISPLAY;

/* Set the internal metrics to the true screen size
 */
   screen_mem.h = screen_buf.h = rows;
   screen_mem.scroll_bottom = screen_buf.scroll_bottom = rows - 1;
   screen_mem.size = screen_buf.size = size;

/* Add the stop byte to speed up vt_refresh_section(). There must
 * be one byte beyond the end which is identical in both buffers
 * and another one which is different.
 */
   screen_mem.img[size+1] = scr_blank;

   return 0;
}

int iniz_screen(int fg, int bg) {


   clrchar  scr_init = { ' ', CLR(fg, bg) };
   int cols, rows;
   get_screen_size( &cols, &rows );

   if( set_screen_size( cols, rows ) )
      return CANNOT_INIT_DISPLAY;

   clrchar mem_blank = { ~0, ~0 };
   int size = rows * cols;

   if ( size < 0 ) return CANNOT_INIT_DISPLAY;

   if( screen_buf.img ) free( screen_buf.img );
   if( screen_mem.img ) free( screen_mem.img );

/* The buffers are made bigger than necessary, so that a stop
 * byte can be added at the end. This improves the loop logic
 * in vt_refresh_section(), so refresh is quicker.
 */
   screen_buf = ct_create( 0, 0, cols, rows + 1, scr_init );
   screen_mem = ct_create( 0, 0, cols, rows + 1, mem_blank );

   if( !screen_buf.img || !screen_mem.img )
      return CANNOT_INIT_DISPLAY;

/* Set the internal metrics to the true screen size
 */
   screen_mem.h = screen_buf.h = rows;
   screen_mem.scroll_bottom = screen_buf.scroll_bottom = rows - 1;
   screen_mem.size = screen_buf.size = size;

/* Add the stop byte to speed up vt_refresh_section(). There must
 * be one byte beyond the end which is identical in both buffers
 * and another one which is different.
 */
   screen_mem.img[size+1] = scr_init;

   return 0;
}

inline void hide_vt_cursor() {

   fwrite( "\033[?25l\033[?1c", sizeof(char), 11, stdout );
}

void hide_text_cursor() {

   hide_vt_cursor();
   fflush( stdout );
   cursor_visible = 0;
}

inline int vt_refresh_strip( int start, int end ) {

   int col = start % screen_buf.w;
   int row = start / screen_buf.w;

   clrchar *next_ch = screen_buf.img + start;
   clrchar *stop_ch = screen_buf.img + end;
   clrchar *mem_ch = screen_mem.img + start;

   put_vt_cursor( col, row );

   while( next_ch != stop_ch ) {

   /* Change output colour, if needed, to match the one for this char
    */
      if( next_ch->clr != output_clr )
         set_output_colour( next_ch->clr );

   /* Use replacement char and colour if control character is output
    */
      if( a_control_char[(unsigned char)next_ch->ch] ) {
         if( a_control_char[(unsigned char)next_ch->ch] == 2 ) 
            enable_box_chars();
         else {
            *next_ch = replacement_char;
            if( next_ch->clr != output_clr )
               set_output_colour( next_ch->clr );
            }
         }

   /* Output the char
    */
      fwrite( &next_ch->ch, sizeof(char), 1, stdout );

   /* Update the memory buffer to reflect the physical screen,
    * and move on to next char
    */
      *mem_ch++ = *next_ch++;
      }

   return 0;
}

int vt_refresh_section( int refresh_start, int refresh_end ) {

   clrchar *mem  = screen_mem.img + refresh_start;
   clrchar *cell = screen_buf.img + refresh_start;
   clrchar *img_end = screen_buf.img + refresh_end;
   int start, end;

   int saved_col = vt_col, saved_row = vt_row;

   while( cell < img_end ) {

      while( *((short*)cell) == *((short*)mem) ) {
         cell++; mem++;
         }

      if( cell >= img_end ) break;

      start = cell - screen_buf.img;

      while( *((short*)cell) != *((short*)mem) ) {
         cell++; mem++;
         }

      if( cell > img_end ) cell = img_end;

      end = cell - screen_buf.img;

      vt_refresh_strip( start, end );
      }

/* Return cursor to its starting position
 */
   put_vt_cursor( saved_col, saved_row );
   fflush( stdout );
   return 0;
}

int vt_refresh() {

   int start = screen_buf.scroll_top * screen_buf.w;
   int end = ( screen_buf.scroll_bottom + 1 ) * screen_buf.w;

   return vt_refresh_section( start, end );
}


int init_display() {

   int cols, rows;
   get_screen_size( &cols, &rows );

   if( set_screen_size( cols, rows ) )
      return CANNOT_INIT_DISPLAY;

   prepare_control_char_lookup();

   hide_text_cursor();
   vt_refresh();

   atexit( restore_display );
   return 0;
}


// ######################################################## FUNZIONI PER LA GESTIONE DEL TESTO COLORATO
int ct_text_xy( ct_image *dst, int x, int y, const char *text, clr_t clr ) {

   int cell, end, row_end;
   const char *ch = text;

   if( !dst || !text || y < 0 || y >= dst->h ) return -1;

   cell = y * dst->w + x;
   end = cell + strlen( text );
   row_end = ( y + 1 ) * dst->w;

/* crop start of line if x < 0 
 */
   if( x < 0 ) {
      cell -= x;
      ch -= x;
      }
      
/* crop end of line if start + length > width 
 */
   if( end > row_end ) end = row_end;
   if( end > dst->size ) end = dst->size;

   for( ; cell < end; cell++, ch++ ) {

      dst->img[cell].ch = *ch;
      dst->img[cell].clr = clr;
      }
   vt_refresh();

   return 0;
}

